from __future__ import annotations

from typing import Any

from strawberry.test.client import BaseGraphQLTestClient, Response


class GraphQLTestClient(BaseGraphQLTestClient):
    async def query(
        self,
        query: str,
        variables: dict[str, Any] | None = None,
        headers: dict[str, object] | None = None,
        files: dict[str, object] | None = None,
        assert_no_errors: bool | None = True,
    ) -> Response:
        body = self._build_body(query, variables, files)

        resp = await self.request(body, headers, files)
        data = await resp.json()

        response = Response(
            errors=data.get("errors"),
            data=data.get("data"),
            extensions=data.get("extensions"),
        )

        if assert_no_errors:
            assert resp.status == 200
            assert response.errors is None

        return response

    async def request(
        self,
        body: dict[str, object],
        headers: dict[str, object] | None = None,
        files: dict[str, object] | None = None,
    ) -> Any:
        return await self._client.post(
            self.url,
            json=body if not files else None,
            data=body if files else None,
        )


__all__ = ["GraphQLTestClient"]



from graphql.validation import NoSchemaIntrospectionCustomRule

from strawberry.extensions import AddValidationRules


class DisableIntrospection(AddValidationRules):
    """Disable introspection queries.

    Example:

    ```python
    import strawberry
    from strawberry.extensions import DisableIntrospection


    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "Hello, world!"


    schema = strawberry.Schema(
        Query,
        extensions=[
            DisableIntrospection(),
        ],
    )
    ```
    """

    def __init__(self) -> None:
        super().__init__([NoSchemaIntrospectionCustomRule])


__all__ = ["DisableIntrospection"]



from __future__ import annotations

import contextlib
import inspect
import types
from asyncio import iscoroutinefunction
from typing import (
    TYPE_CHECKING,
    Any,
    NamedTuple,
)

from strawberry.extensions import SchemaExtension

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable, Iterator
    from types import TracebackType

    from strawberry.extensions.base_extension import Hook
    from strawberry.types import StreamExecutionResult
    from strawberry.utils.await_maybe import AwaitableOrValue


class WrappedHook(NamedTuple):
    extension: SchemaExtension
    hook: Callable[
        ...,
        contextlib.AbstractAsyncContextManager[None]
        | contextlib.AbstractContextManager[None],
    ]
    is_async: bool


class ExtensionContextManagerBase:
    __slots__ = (
        "async_exit_stack",
        "default_hook",
        "exit_stack",
        "hooks",
    )

    HOOK_NAME: str

    def __init__(self, extensions: list[SchemaExtension]) -> None:
        self.hooks: list[WrappedHook] = []
        self.default_hook: Hook = getattr(SchemaExtension, self.HOOK_NAME)
        for extension in extensions:
            hook = self.get_hook(extension)
            if hook:
                self.hooks.append(hook)

    def get_hook(self, extension: SchemaExtension) -> WrappedHook | None:
        hook_fn: Hook | None = getattr(type(extension), self.HOOK_NAME)
        hook_fn = hook_fn if hook_fn is not self.default_hook else None

        if hook_fn:
            if inspect.isgeneratorfunction(hook_fn):
                context_manager = contextlib.contextmanager(
                    types.MethodType(hook_fn, extension)
                )
                return WrappedHook(
                    extension=extension, hook=context_manager, is_async=False
                )

            if inspect.isasyncgenfunction(hook_fn):
                context_manager_async = contextlib.asynccontextmanager(
                    types.MethodType(hook_fn, extension)
                )
                return WrappedHook(
                    extension=extension, hook=context_manager_async, is_async=True
                )

            if callable(hook_fn):
                return self.from_callable(extension, hook_fn)  # type: ignore[arg-type]

            raise ValueError(
                f"Hook {self.HOOK_NAME} on {extension} "
                f"must be callable, received {hook_fn!r}"
            )

        return None  # Current extension does not define a hook for this lifecycle stage

    @staticmethod
    def from_callable(
        extension: SchemaExtension,
        func: Callable[[SchemaExtension], AwaitableOrValue[Any]],
    ) -> WrappedHook:
        if iscoroutinefunction(func):

            @contextlib.asynccontextmanager
            async def iterator() -> AsyncIterator[None]:
                await func(extension)
                yield

            return WrappedHook(extension=extension, hook=iterator, is_async=True)

        @contextlib.contextmanager  # type: ignore[no-redef]
        def iterator() -> Iterator[None]:
            func(extension)
            yield

        return WrappedHook(extension=extension, hook=iterator, is_async=False)

    def __enter__(self) -> None:
        self.exit_stack = contextlib.ExitStack()

        self.exit_stack.__enter__()

        for hook in self.hooks:
            if hook.is_async:
                raise RuntimeError(
                    f"SchemaExtension hook {hook.extension}.{self.HOOK_NAME} "
                    "failed to complete synchronously."
                )
            self.exit_stack.enter_context(hook.hook())  # type: ignore

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.exit_stack.__exit__(exc_type, exc_val, exc_tb)

    async def __aenter__(self) -> None:
        self.async_exit_stack = contextlib.AsyncExitStack()

        await self.async_exit_stack.__aenter__()

        for hook in self.hooks:
            if hook.is_async:
                await self.async_exit_stack.enter_async_context(hook.hook())  # type: ignore
            else:
                self.async_exit_stack.enter_context(hook.hook())  # type: ignore

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.async_exit_stack.__aexit__(exc_type, exc_val, exc_tb)


class OperationContextManager(ExtensionContextManagerBase):
    HOOK_NAME = SchemaExtension.on_operation.__name__


class ValidationContextManager(ExtensionContextManagerBase):
    HOOK_NAME = SchemaExtension.on_validate.__name__


class ParsingContextManager(ExtensionContextManagerBase):
    HOOK_NAME = SchemaExtension.on_parse.__name__


class ExecutingContextManager(ExtensionContextManagerBase):
    HOOK_NAME = SchemaExtension.on_execute.__name__


class StreamResultContextManager(ExtensionContextManagerBase):
    HOOK_NAME = SchemaExtension.on_stream_result.__name__

    def __init__(
        self, extensions: list[SchemaExtension], result: StreamExecutionResult
    ) -> None:
        self.result = result
        super().__init__(extensions)

    def from_callable(  # type: ignore[override]
        self,
        extension: SchemaExtension,
        func: Callable[..., AwaitableOrValue[Any]],
    ) -> WrappedHook:
        if iscoroutinefunction(func):

            @contextlib.asynccontextmanager
            async def async_iterator(
                result: StreamExecutionResult,
            ) -> AsyncIterator[None]:
                await func(extension, result)
                yield

            return WrappedHook(extension=extension, hook=async_iterator, is_async=True)

        @contextlib.contextmanager
        def sync_iterator(result: StreamExecutionResult) -> Iterator[None]:
            func(extension, result)
            yield

        return WrappedHook(extension=extension, hook=sync_iterator, is_async=False)

    def __enter__(self) -> None:
        self.exit_stack = contextlib.ExitStack()
        self.exit_stack.__enter__()

        for hook in self.hooks:
            if hook.is_async:
                raise RuntimeError(
                    f"SchemaExtension hook {hook.extension}.{self.HOOK_NAME} "
                    "failed to complete synchronously."
                )
            self.exit_stack.enter_context(hook.hook(self.result))  # type: ignore

    async def __aenter__(self) -> None:
        self.async_exit_stack = contextlib.AsyncExitStack()
        await self.async_exit_stack.__aenter__()

        for hook in self.hooks:
            if hook.is_async:
                await self.async_exit_stack.enter_async_context(hook.hook(self.result))  # type: ignore
            else:
                self.async_exit_stack.enter_context(hook.hook(self.result))  # type: ignore



from typing import Any, TypeAlias, Union

try:
    from graphql.execution import ExecutionContext as GraphQLExecutionContext

    _execution_context_class_arg = "execution_context_class"
except ImportError:
    from graphql.execution import (  # type: ignore[attr-defined,no-redef]
        Executor as GraphQLExecutionContext,  # pyright: ignore[reportAttributeAccessIssue]
    )

    _execution_context_class_arg = "executor_class"

from graphql.execution import ExecutionResult as OriginalGraphQLExecutionResult
from graphql.execution import execute, subscribe

from strawberry.types import ExecutionResult

try:
    from graphql import (  # type: ignore[attr-defined]
        ExperimentalIncrementalExecutionResults as GraphQLIncrementalExecutionResults,  # pyright: ignore[reportAttributeAccessIssue]
    )
    from graphql.execution import (  # type: ignore[attr-defined]
        InitialIncrementalExecutionResult,  # pyright: ignore[reportAttributeAccessIssue]
        SubsequentIncrementalExecutionResult,  # pyright: ignore[reportAttributeAccessIssue]
        experimental_execute_incrementally,  # pyright: ignore[reportAttributeAccessIssue]
    )
    from graphql.type.directives import (  # type: ignore[attr-defined]
        GraphQLDeferDirective,  # pyright: ignore[reportAttributeAccessIssue]
        GraphQLStreamDirective,  # pyright: ignore[reportAttributeAccessIssue]
    )

    incremental_execution_directives = (
        GraphQLDeferDirective,
        GraphQLStreamDirective,
    )

    GraphQLExecutionResult: TypeAlias = (
        OriginalGraphQLExecutionResult | InitialIncrementalExecutionResult
    )

    # The individual frames produced when an incremental delivery container
    # (`@defer`/`@stream`) is expanded into a flat stream of results.
    GraphQLIncrementalResult: TypeAlias = (
        InitialIncrementalExecutionResult | SubsequentIncrementalExecutionResult
    )

except ImportError:

    class GraphQLIncrementalExecutionResults:  # type: ignore[no-redef]
        pass

    class InitialIncrementalExecutionResult:  # type: ignore[no-redef]
        pass

    class SubsequentIncrementalExecutionResult:  # type: ignore[no-redef]
        pass

    GraphQLExecutionResult = OriginalGraphQLExecutionResult  # type: ignore
    # Incremental delivery isn't available on graphql-core < 3.3, so no frame
    # type exists. Fall back to the (empty) container type so annotations remain
    # importable at runtime.
    GraphQLIncrementalResult = GraphQLIncrementalExecutionResults  # type: ignore

    incremental_execution_directives = ()  # type: ignore
    experimental_execute_incrementally = None


def execution_context_class_kwargs(
    execution_context_class: type[GraphQLExecutionContext] | None,
) -> dict[str, Any]:
    if execution_context_class is None:
        return {}

    return {_execution_context_class_arg: execution_context_class}


# TODO: give this a better name, maybe also a better place
ResultType = Union[  # noqa: UP007
    OriginalGraphQLExecutionResult,
    GraphQLIncrementalExecutionResults,
    ExecutionResult,
]

__all__ = [
    "GraphQLExecutionContext",
    "GraphQLExecutionResult",
    "GraphQLIncrementalExecutionResults",
    "GraphQLIncrementalResult",
    "InitialIncrementalExecutionResult",
    "ResultType",
    "SubsequentIncrementalExecutionResult",
    "execute",
    "execution_context_class_kwargs",
    "experimental_execute_incrementally",
    "incremental_execution_directives",
    "subscribe",
]



from collections.abc import Callable, Iterator
from typing import Protocol, runtime_checkable

from graphql import ExecutionResult as GraphQLExecutionResult
from graphql.error import GraphQLError

from strawberry.extensions.base_extension import SchemaExtension
from strawberry.types.execution import (
    ExecutionResult as StrawberryExecutionResult,
)
from strawberry.types.execution import StreamExecutionResult


@runtime_checkable
class _ResultWithErrors(Protocol):
    errors: list[GraphQLError] | None


def default_should_mask_error(_: GraphQLError) -> bool:
    # Mask all errors
    return True


class MaskErrors(SchemaExtension):
    should_mask_error: Callable[[GraphQLError], bool]
    error_message: str

    def __init__(
        self,
        should_mask_error: Callable[[GraphQLError], bool] = default_should_mask_error,
        error_message: str = "Unexpected error.",
    ) -> None:
        self.should_mask_error = should_mask_error
        self.error_message = error_message
        self._stream_result_processed = False

    def anonymise_error(self, error: GraphQLError) -> GraphQLError:
        return GraphQLError(
            message=self.error_message,
            nodes=error.nodes,
            source=error.source,
            positions=error.positions,
            path=error.path,
            original_error=None,
        )

    def _process_errors(self, errors: list[GraphQLError]) -> list[GraphQLError]:
        processed_errors: list[GraphQLError] = []

        for error in errors:
            if self.should_mask_error(error):
                processed_errors.append(self.anonymise_error(error))
            else:
                processed_errors.append(error)

        return processed_errors

    def _process_result(self, result: object) -> None:
        if isinstance(result, _ResultWithErrors) and result.errors:
            result.errors = self._process_errors(result.errors)

    def _process_stream_result(self, result: StreamExecutionResult) -> None:
        self._process_result(result)

        for incremental_result in getattr(result, "incremental", None) or ():
            self._process_result(incremental_result)

        for completed_result in getattr(result, "completed", None) or ():
            self._process_result(completed_result)

    def on_operation(self) -> Iterator[None]:
        self._stream_result_processed = False
        yield

        # Streaming operations are handled result-by-result before each frame is
        # yielded. Avoid processing the last result again when the stream closes.
        if self._stream_result_processed:
            return

        result = self.execution_context.result

        if isinstance(result, (GraphQLExecutionResult, StrawberryExecutionResult)):
            self._process_result(result)
        elif initial_result := getattr(result, "initial_result", None):
            self._process_result(initial_result)

    def on_stream_result(self, result: StreamExecutionResult) -> Iterator[None]:
        """Mask errors before a streamed execution result reaches the client."""
        self._stream_result_processed = True
        self._process_stream_result(result)
        yield None



from __future__ import annotations

import dataclasses
from typing import (
    TYPE_CHECKING,
    Any,
    TypeAlias,
    runtime_checkable,
)
from typing_extensions import NotRequired, Protocol, TypedDict

from graphql import specified_rules

try:
    from graphql.execution import (  # type: ignore[attr-defined]
        InitialIncrementalExecutionResult as _InitialIncrementalExecutionResult,  # pyright: ignore[reportAttributeAccessIssue]
    )
    from graphql.execution import (  # type: ignore[attr-defined]
        SubsequentIncrementalExecutionResult as _SubsequentIncrementalExecutionResult,  # pyright: ignore[reportAttributeAccessIssue]
    )
except ImportError:

    class _InitialIncrementalExecutionResult:  # type: ignore[no-redef]
        extensions: dict[str, Any] | None

    class _SubsequentIncrementalExecutionResult:  # type: ignore[no-redef]
        extensions: dict[str, Any] | None


from strawberry.utils.operation import get_first_operation, get_operation_type

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing_extensions import NotRequired

    from graphql import ASTValidationRule
    from graphql.error.graphql_error import GraphQLError
    from graphql.language import DocumentNode, OperationDefinitionNode

    from strawberry.schema import Schema
    from strawberry.schema._graphql_core import GraphQLExecutionResult

    from .graphql import OperationType


class ParseOptions(TypedDict):
    max_tokens: NotRequired[int]


@dataclasses.dataclass
class ExecutionContext:
    query: str | None
    schema: Schema
    allowed_operations: Iterable[OperationType]
    context: Any = None
    variables: dict[str, Any] | None = None
    parse_options: ParseOptions = dataclasses.field(default_factory=ParseOptions)
    root_value: Any | None = None
    validation_rules: tuple[type[ASTValidationRule], ...] = dataclasses.field(
        default_factory=lambda: tuple(specified_rules)
    )

    # The operation name that is provided by the request
    provided_operation_name: dataclasses.InitVar[str | None] = None

    # Values that get populated during the GraphQL execution so that they can be
    # accessed by extensions
    graphql_document: DocumentNode | None = None
    pre_execution_errors: list[GraphQLError] | None = None
    result: GraphQLExecutionResult | None = None
    extensions_results: dict[str, Any] = dataclasses.field(default_factory=dict)

    operation_extensions: dict[str, Any] | None = None

    def __post_init__(self, provided_operation_name: str | None) -> None:
        self._provided_operation_name = provided_operation_name

    @property
    def operation_name(self) -> str | None:
        if self._provided_operation_name is not None:
            return self._provided_operation_name

        definition = self._get_first_operation()
        if not definition:
            return None

        if not definition.name:
            return None

        return definition.name.value

    @property
    def operation_type(self) -> OperationType:
        graphql_document = self.graphql_document
        if not graphql_document:
            raise RuntimeError("No GraphQL document available")

        return get_operation_type(graphql_document, self.operation_name)

    def _get_first_operation(self) -> OperationDefinitionNode | None:
        graphql_document = self.graphql_document
        if not graphql_document:
            return None

        return get_first_operation(graphql_document)


@dataclasses.dataclass
class ExecutionResult:
    data: dict[str, Any] | None
    errors: list[GraphQLError] | None
    extensions: dict[str, Any] | None = None


@dataclasses.dataclass
class PreExecutionError(ExecutionResult):
    """Differentiate between a normal execution result and an immediate error.

    Immediate errors are errors that occur before the execution phase i.e validation errors,
    or any other error that occur before we interact with resolvers.

    These errors are required by `graphql-ws-transport` protocol in order to close the operation
    right away once the error is encountered.
    """


StreamExecutionResult: TypeAlias = (
    ExecutionResult
    | _InitialIncrementalExecutionResult
    | _SubsequentIncrementalExecutionResult
)


@runtime_checkable
class SubscriptionExecutionResult(Protocol):
    def __aiter__(self) -> SubscriptionExecutionResult:  # pragma: no cover
        ...

    async def __anext__(self) -> Any:  # pragma: no cover
        ...


__all__ = [
    "ExecutionContext",
    "ExecutionResult",
    "ParseOptions",
    "StreamExecutionResult",
    "SubscriptionExecutionResult",
]



from .annotated_unions import ConvertUnionToAnnotatedUnion
from .maybe_optional import ConvertMaybeToOptional
from .replace_scalar_wrappers import ReplaceScalarWrappers
from .update_imports import UpdateImportsCodemod

__all__ = [
    "ConvertMaybeToOptional",
    "ConvertUnionToAnnotatedUnion",
    "ReplaceScalarWrappers",
    "UpdateImportsCodemod",
]



from .argument import argument
from .enum import enum, enum_value
from .field import field
from .mutation import mutation
from .object_type import input, interface, interface_object, type  # noqa: A004
from .scalar import scalar
from .schema import Schema
from .schema_directive import schema_directive
from .union import union

__all__ = [
    "Schema",
    "argument",
    "enum",
    "enum_value",
    "field",
    "input",
    "interface",
    "interface_object",
    "mutation",
    "scalar",
    "schema_directive",
    "type",
    "union",
]



from typing import Any

from graphql import (
    ArgumentNode,
    GraphQLError,
    GraphQLNamedType,
    ObjectValueNode,
    ValidationContext,
    ValidationRule,
    get_named_type,
)

from strawberry.types.base import StrawberryMaybe, StrawberryOptional
from strawberry.utils.str_converters import to_camel_case


class MaybeNullValidationRule(ValidationRule):
    """Validates that Maybe[T] fields do not receive explicit null values.

    This rule ensures that:
    - Maybe[T] fields can only be omitted or have non-null values
    - Maybe[T | None] fields can be omitted, null, or have non-null values

    This provides clear semantics where Maybe[T] means "either present with value or absent"
    and Maybe[T | None] means "present with value, present but null, or absent".
    """

    def __init__(self, validation_context: ValidationContext) -> None:
        super().__init__(validation_context)

    def enter_argument(self, node: ArgumentNode, *_args: Any) -> None:
        # Check if this is a null value
        if node.value.kind != "null_value":
            return

        # Get the argument definition from the schema
        argument_def = self.context.get_argument()
        if not argument_def:
            return

        # Check if this argument corresponds to a Maybe[T] (not Maybe[T | None])
        # The argument type extensions should contain the Strawberry type info
        strawberry_arg_info = argument_def.extensions.get("strawberry-definition")
        if not strawberry_arg_info:
            return

        # Get the Strawberry type from the argument info
        field_type = getattr(strawberry_arg_info, "type", None)
        if not field_type:
            return

        if isinstance(field_type, StrawberryMaybe) and not isinstance(
            field_type.of_type, StrawberryOptional
        ):
            # This is Maybe[T] - should not accept null values
            type_name = self._get_type_name(field_type.of_type)

            self.report_error(
                GraphQLError(
                    f"Expected value of type '{type_name}', found null. "
                    f"Argument '{node.name.value}' of type 'Maybe[{type_name}]' cannot be explicitly set to null. "
                    f"Use 'Maybe[{type_name} | None]' if you need to allow null values.",
                    nodes=[node],
                )
            )

    def enter_object_value(self, node: ObjectValueNode, *_args: Any) -> None:
        # Get the input type for this object
        input_type = get_named_type(self.context.get_input_type())
        if not input_type:
            return

        # Get the Strawberry type definition from extensions
        strawberry_type = input_type.extensions.get("strawberry-definition")
        if not strawberry_type:
            return

        # Check each field in the object for null Maybe[T] violations
        self.validate_maybe_fields(node, input_type, strawberry_type)

    def validate_maybe_fields(
        self, node: ObjectValueNode, input_type: GraphQLNamedType, strawberry_type: Any
    ) -> None:
        # Create a map of field names to field nodes for easy lookup
        field_node_map = {field.name.value: field for field in node.fields}

        # Check each field in the Strawberry type definition
        if not hasattr(strawberry_type, "fields"):
            return

        for field_def in strawberry_type.fields:
            # Resolve the actual GraphQL field name using the same logic as NameConverter
            if field_def.graphql_name is not None:
                field_name = field_def.graphql_name
            else:
                # Apply auto_camel_case conversion if enabled (default behavior)
                field_name = to_camel_case(field_def.python_name)

            # Check if this field is present in the input and has a null value
            if field_name in field_node_map:
                field_node = field_node_map[field_name]

                # Check if this field has a null value
                if field_node.value.kind == "null_value":
                    # Check if this is a Maybe[T] (not Maybe[T | None])
                    field_type = field_def.type
                    if isinstance(field_type, StrawberryMaybe) and not isinstance(
                        field_type.of_type, StrawberryOptional
                    ):
                        # This is Maybe[T] - should not accept null values
                        type_name = self._get_type_name(field_type.of_type)
                        self.report_error(
                            GraphQLError(
                                f"Expected value of type '{type_name}', found null. "
                                f"Field '{field_name}' of type 'Maybe[{type_name}]' cannot be explicitly set to null. "
                                f"Use 'Maybe[{type_name} | None]' if you need to allow null values.",
                                nodes=[field_node],
                            )
                        )

    def _get_type_name(self, type_: Any) -> str:
        """Get a readable type name for error messages."""
        if hasattr(type_, "__name__"):
            return type_.__name__
        # Handle Strawberry types that don't have __name__
        if hasattr(type_, "of_type") and hasattr(type_.of_type, "__name__"):
            # For StrawberryList, StrawberryOptional, etc.
            return (
                f"list[{type_.of_type.__name__}]"
                if "List" in str(type_.__class__)
                else type_.of_type.__name__
            )
        return str(type_)


__all__ = ["MaybeNullValidationRule"]



from __future__ import annotations

import libcst as cst
import libcst.matchers as m
from libcst._nodes.expression import BaseExpression  # noqa: TC002
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.codemod.visitors import AddImportsVisitor


class ConvertMaybeToOptional(VisitorBasedCodemodCommand):
    DESCRIPTION: str = (
        "Converts strawberry.Maybe[T] to strawberry.Maybe[T | None] "
        "to match the new Maybe type definition"
    )

    def __init__(
        self,
        context: CodemodContext,
        use_pipe_syntax: bool = True,  # Default to pipe syntax for modern Python
    ) -> None:
        self.use_pipe_syntax = use_pipe_syntax
        super().__init__(context)

    @m.leave(
        m.Subscript(
            value=m.Attribute(value=m.Name("strawberry"), attr=m.Name("Maybe"))
            | m.Name("Maybe")
        )
    )
    def leave_maybe_subscript(
        self, original_node: cst.Subscript, updated_node: cst.Subscript
    ) -> BaseExpression:
        # Skip if it's not a strawberry.Maybe or imported Maybe
        if isinstance(original_node.value, cst.Name):
            # Check if this is an imported Maybe from strawberry
            # For now, we'll assume any standalone "Maybe" is from strawberry
            # In a more robust implementation, we'd track imports
            pass

        # Get the inner type
        if isinstance(original_node.slice, (list, tuple)):
            if len(original_node.slice) != 1:
                return original_node
            slice_element = original_node.slice[0]
        else:
            slice_element = original_node.slice

        if not isinstance(slice_element, cst.SubscriptElement):
            return original_node

        if not isinstance(slice_element.slice, cst.Index):
            return original_node

        inner_type = slice_element.slice.value

        # Check if the inner type already includes None
        if self._already_includes_none(inner_type):
            return original_node

        # Create the new union type with None
        new_type: BaseExpression
        if self.use_pipe_syntax:
            new_type = cst.BinaryOperation(
                left=inner_type,
                operator=cst.BitOr(
                    whitespace_before=cst.SimpleWhitespace(" "),
                    whitespace_after=cst.SimpleWhitespace(" "),
                ),
                right=cst.Name("None"),
            )
        else:
            # Use Union[T, None] syntax
            AddImportsVisitor.add_needed_import(self.context, "typing", "Union")
            new_type = cst.Subscript(
                value=cst.Name("Union"),
                slice=[
                    cst.SubscriptElement(slice=cst.Index(value=inner_type)),
                    cst.SubscriptElement(slice=cst.Index(value=cst.Name("None"))),
                ],
            )

        # Return the updated Maybe[T | None]
        return updated_node.with_changes(
            slice=[cst.SubscriptElement(slice=cst.Index(value=new_type))]
        )

    def _already_includes_none(self, node: BaseExpression) -> bool:
        """Check if the type already includes None (e.g., T | None or Union[T, None])."""
        # Check for T | None pattern
        if isinstance(node, cst.BinaryOperation) and isinstance(
            node.operator, cst.BitOr
        ):
            if isinstance(node.right, cst.Name) and node.right.value == "None":
                return True
            # Recursively check left side for chained unions
            if self._already_includes_none(node.left):
                return True

        # Check for Union[..., None] pattern
        if (
            isinstance(node, cst.Subscript)
            and isinstance(node.value, cst.Name)
            and node.value.value == "Union"
        ):
            # Handle both list and tuple slice formats
            slice_elements = (
                node.slice if isinstance(node.slice, (list, tuple)) else [node.slice]
            )
            for element in slice_elements:
                if (
                    isinstance(element, cst.SubscriptElement)
                    and isinstance(element.slice, cst.Index)
                    and isinstance(element.slice.value, cst.Name)
                    and element.slice.value.value == "None"
                ):
                    return True

        return False



GRAPHQL_TRANSPORT_WS_PROTOCOL = "graphql-transport-ws"
GRAPHQL_WS_PROTOCOL = "graphql-ws"
GRAPHQL_SSE_PROTOCOL = "graphql-sse"
MULTIPART_SUBSCRIPTION_PROTOCOL = "multipart-subscription"


__all__ = [
    "GRAPHQL_SSE_PROTOCOL",
    "GRAPHQL_TRANSPORT_WS_PROTOCOL",
    "GRAPHQL_WS_PROTOCOL",
    "MULTIPART_SUBSCRIPTION_PROTOCOL",
]



from collections.abc import Callable, Iterable
from typing import (
    Any,
    NewType,
    TypeVar,
    overload,
)

from graphql import (
    GraphQLScalarLiteralParser,
    GraphQLScalarSerializer,
    GraphQLScalarValueParser,
)

from strawberry.types.scalar import ScalarWrapper, _process_scalar

_T = TypeVar("_T", bound=type | NewType)


def identity(x: _T) -> _T:  # pragma: no cover
    return x


@overload
def scalar(
    *,
    name: str | None = None,
    description: str | None = None,
    specified_by_url: str | None = None,
    serialize: GraphQLScalarSerializer = identity,
    parse_value: GraphQLScalarValueParser | None = None,
    parse_literal: GraphQLScalarLiteralParser | None = None,
    directives: Iterable[object] = (),
    authenticated: bool = False,
    inaccessible: bool = False,
    policy: list[list[str]] | None = None,
    requires_scopes: list[list[str]] | None = None,
    tags: Iterable[str] | None = (),
) -> Callable[[_T], _T]: ...


@overload
def scalar(
    cls: _T,
    *,
    name: str | None = None,
    description: str | None = None,
    specified_by_url: str | None = None,
    serialize: GraphQLScalarSerializer = identity,
    parse_value: GraphQLScalarValueParser | None = None,
    parse_literal: GraphQLScalarLiteralParser | None = None,
    directives: Iterable[object] = (),
    authenticated: bool = False,
    inaccessible: bool = False,
    policy: list[list[str]] | None = None,
    requires_scopes: list[list[str]] | None = None,
    tags: Iterable[str] | None = (),
) -> _T: ...


def scalar(
    cls: _T | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    specified_by_url: str | None = None,
    serialize: GraphQLScalarSerializer = identity,
    parse_value: GraphQLScalarValueParser | None = None,
    parse_literal: GraphQLScalarLiteralParser | None = None,
    directives: Iterable[object] = (),
    authenticated: bool = False,
    inaccessible: bool = False,
    policy: list[list[str]] | None = None,
    requires_scopes: list[list[str]] | None = None,
    tags: Iterable[str] | None = (),
) -> Any:
    """Annotates a class or type as a GraphQL custom scalar.

    Args:
        cls: The class or type to annotate
        name: The GraphQL name of the scalar
        description: The description of the scalar
        specified_by_url: The URL of the specification
        serialize: The function to serialize the scalar
        parse_value: The function to parse the value
        parse_literal: The function to parse the literal
        directives: The directives to apply to the scalar
        authenticated: Whether to add the @authenticated directive
        inaccessible: Whether to add the @inaccessible directive
        policy: The list of policy names to add to the @policy directive
        requires_scopes: The list of scopes to add to the @requires directive
        tags: The list of tags to add to the @tag directive

    Returns:
        The decorated class or type

    Example usages:

    ```python
    strawberry.federation.scalar(
        datetime.date,
        serialize=lambda value: value.isoformat(),
        parse_value=datetime.parse_date,
    )

    Base64Encoded = strawberry.federation.scalar(
        NewType("Base64Encoded", bytes),
        serialize=base64.b64encode,
        parse_value=base64.b64decode,
    )


    @strawberry.federation.scalar(
        serialize=lambda value: ",".join(value.items),
        parse_value=lambda value: CustomList(value.split(",")),
    )
    class CustomList:
        def __init__(self, items):
            self.items = items
    ```
    """
    from strawberry.federation.schema_directives import (
        Authenticated,
        Inaccessible,
        Policy,
        RequiresScopes,
        Tag,
    )

    if parse_value is None:
        parse_value = cls

    directives = list(directives)

    if authenticated:
        directives.append(Authenticated())

    if inaccessible:
        directives.append(Inaccessible())

    if policy:
        directives.append(Policy(policies=policy))

    if requires_scopes:
        directives.append(RequiresScopes(scopes=requires_scopes))

    if tags:
        directives.extend(Tag(name=tag) for tag in tags)

    def wrap(cls: _T) -> ScalarWrapper:
        return _process_scalar(
            cls,
            name=name,
            description=description,
            specified_by_url=specified_by_url,
            serialize=serialize,
            parse_value=parse_value,
            parse_literal=parse_literal,
            directives=directives,
        )

    if cls is None:
        return wrap

    return wrap(cls)


__all__ = ["scalar"]



from collections.abc import Iterable
from typing import Any

from strawberry.types.arguments import StrawberryArgumentAnnotation


def argument(
    description: str | None = None,
    name: str | None = None,
    deprecation_reason: str | None = None,
    directives: Iterable[object] = (),
    inaccessible: bool = False,
    tags: Iterable[str] | None = (),
    graphql_type: Any | None = None,
) -> StrawberryArgumentAnnotation:
    from strawberry.federation.schema_directives import Inaccessible, Tag

    directives = list(directives)

    if inaccessible:
        directives.append(Inaccessible())

    if tags:
        directives.extend(Tag(name=tag) for tag in tags)

    return StrawberryArgumentAnnotation(
        description=description,
        name=name,
        deprecation_reason=deprecation_reason,
        directives=directives,
        graphql_type=graphql_type,
    )


__all__ = ["argument"]



import importlib
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .apollo import ApolloTracingExtension, ApolloTracingExtensionSync
    from .apollo_federation import (
        ApolloFederationTracingExtension,
        ApolloFederationTracingExtensionSync,
    )
    from .datadog import DatadogTracingExtension, DatadogTracingExtensionSync
    from .opentelemetry import (
        OpenTelemetryExtension,
        OpenTelemetryExtensionSync,
    )

__all__ = [
    "ApolloFederationTracingExtension",
    "ApolloFederationTracingExtensionSync",
    "ApolloTracingExtension",
    "ApolloTracingExtensionSync",
    "DatadogTracingExtension",
    "DatadogTracingExtensionSync",
    "OpenTelemetryExtension",
    "OpenTelemetryExtensionSync",
]


def __getattr__(name: str) -> Any:
    if name in {
        "ApolloFederationTracingExtension",
        "ApolloFederationTracingExtensionSync",
    }:
        return getattr(importlib.import_module(".apollo_federation", __name__), name)

    if name in {"DatadogTracingExtension", "DatadogTracingExtensionSync"}:
        return getattr(importlib.import_module(".datadog", __name__), name)

    if name in {"ApolloTracingExtension", "ApolloTracingExtensionSync"}:
        return getattr(importlib.import_module(".apollo", __name__), name)

    if name in {"OpenTelemetryExtension", "OpenTelemetryExtensionSync"}:
        return getattr(importlib.import_module(".opentelemetry", __name__), name)

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from __future__ import annotations

from collections.abc import Callable
from enum import Enum
from typing import TYPE_CHECKING, Any

from strawberry.types import StreamExecutionResult
from strawberry.utils.await_maybe import AsyncIteratorOrIterator, AwaitableOrValue

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from strawberry.types import ExecutionContext


class LifecycleStep(Enum):
    OPERATION = "operation"
    VALIDATION = "validation"
    PARSE = "parse"
    RESOLVE = "resolve"


class SchemaExtension:
    execution_context: ExecutionContext

    # to support extensions that still use the old signature
    # we have an optional argument here for ease of initialization.
    def __init__(
        self, *, execution_context: ExecutionContext | None = None
    ) -> None: ...
    def on_operation(  # type: ignore
        self,
    ) -> AsyncIteratorOrIterator[None]:  # pragma: no cover
        """Called before and after a GraphQL operation (query / mutation) starts."""
        yield None

    def on_validate(  # type: ignore
        self,
    ) -> AsyncIteratorOrIterator[None]:  # pragma: no cover
        """Called before and after the validation step."""
        yield None

    def on_parse(  # type: ignore
        self,
    ) -> AsyncIteratorOrIterator[None]:  # pragma: no cover
        """Called before and after the parsing step."""
        yield None

    def on_execute(  # type: ignore
        self,
    ) -> AsyncIteratorOrIterator[None]:  # pragma: no cover
        """Called before and after the execution step."""
        yield None

    def on_stream_result(  # type: ignore
        self, result: StreamExecutionResult
    ) -> AsyncIteratorOrIterator[None]:  # pragma: no cover
        """Called before and after each result yielded by ``Schema.stream``.

        Extensions can mutate ``result`` before yielding to change the response sent
        by streaming transports. Incremental-delivery frames are included.
        """
        yield None

    def resolve(
        self,
        _next: Callable,
        root: Any,
        info: GraphQLResolveInfo,
        *args: str,
        **kwargs: Any,
    ) -> AwaitableOrValue[object]:
        return _next(root, info, *args, **kwargs)

    def get_results(self) -> AwaitableOrValue[dict[str, Any]]:
        return {}

    @classmethod
    def _implements_resolve(cls) -> bool:
        """Whether the extension implements the resolve method."""
        return cls.resolve is not SchemaExtension.resolve


Hook = (
    Callable[[SchemaExtension], AsyncIteratorOrIterator[None]]
    | Callable[[SchemaExtension, StreamExecutionResult], AsyncIteratorOrIterator[None]]
)

HOOK_METHODS: set[str] = {
    SchemaExtension.on_operation.__name__,
    SchemaExtension.on_validate.__name__,
    SchemaExtension.on_parse.__name__,
    SchemaExtension.on_execute.__name__,
    SchemaExtension.on_stream_result.__name__,
}

__all__ = ["HOOK_METHODS", "Hook", "LifecycleStep", "SchemaExtension"]



from __future__ import annotations

import uuid
from typing import (
    TYPE_CHECKING,
    Any,
)

from channels.testing.websocket import WebsocketCommunicator
from graphql import GraphQLError, GraphQLFormattedError

from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL
from strawberry.subscriptions.protocols.graphql_transport_ws import (
    types as transport_ws_types,
)
from strawberry.subscriptions.protocols.graphql_ws import types as ws_types
from strawberry.types import ExecutionResult

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from types import TracebackType
    from typing_extensions import Self

    from asgiref.typing import ASGIApplication


class GraphQLWebsocketCommunicator(WebsocketCommunicator):
    """A test communicator for GraphQL over Websockets.

    ```python
    import pytest
    from strawberry.channels.testing import GraphQLWebsocketCommunicator
    from myapp.asgi import application


    @pytest.fixture
    async def gql_communicator():
        async with GraphQLWebsocketCommunicator(application, path="/graphql") as client:
            yield client


    async def test_subscribe_echo(gql_communicator):
        async for res in gql_communicator.subscribe(
            query='subscription { echo(message: "Hi") }'
        ):
            assert res.data == {"echo": "Hi"}
    ```
    """

    def __init__(
        self,
        application: ASGIApplication,
        path: str,
        headers: list[tuple[bytes, bytes]] | None = None,
        protocol: str = GRAPHQL_TRANSPORT_WS_PROTOCOL,
        connection_params: dict | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a new communicator.

        Args:
            application: Your asgi application that encapsulates the strawberry schema.
            path: the url endpoint for the schema.
            protocol: currently this supports `graphql-transport-ws` only.
            connection_params: a dictionary of connection parameters to send to the server.
            headers: a list of tuples to be sent as headers to the server.
            subprotocols: an ordered list of preferred subprotocols to be sent to the server.
            **kwargs: additional arguments to be passed to the `WebsocketCommunicator` constructor.
        """
        if connection_params is None:
            connection_params = {}
        self.protocol = protocol
        subprotocols = kwargs.get("subprotocols", [])
        subprotocols.append(protocol)
        self.connection_params = connection_params
        super().__init__(application, path, headers, subprotocols=subprotocols)

    async def __aenter__(self) -> Self:
        await self.gql_init()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.disconnect()

    async def gql_init(self) -> None:
        res = await self.connect()
        if self.protocol == GRAPHQL_TRANSPORT_WS_PROTOCOL:
            assert res == (True, GRAPHQL_TRANSPORT_WS_PROTOCOL)
            await self.send_json_to(
                transport_ws_types.ConnectionInitMessage(
                    {"type": "connection_init", "payload": self.connection_params}
                )
            )
            transport_ws_connection_ack_message: transport_ws_types.ConnectionAckMessage = await self.receive_json_from()
            assert transport_ws_connection_ack_message == {"type": "connection_ack"}
        else:
            assert res == (True, GRAPHQL_WS_PROTOCOL)
            await self.send_json_to(
                ws_types.ConnectionInitMessage({"type": "connection_init"})
            )
            ws_connection_ack_message: ws_types.ConnectionAckMessage = (
                await self.receive_json_from()
            )
            assert ws_connection_ack_message["type"] == "connection_ack"

    # Actual `ExecutionResult`` objects are not available client-side, since they
    # get transformed into `FormattedExecutionResult` on the wire, but we attempt
    # to do a limited representation of them here, to make testing simpler.
    async def subscribe(
        self, query: str, variables: dict | None = None
    ) -> ExecutionResult | AsyncIterator[ExecutionResult]:
        id_ = uuid.uuid4().hex

        if self.protocol == GRAPHQL_TRANSPORT_WS_PROTOCOL:
            await self.send_json_to(
                transport_ws_types.SubscribeMessage(
                    {
                        "id": id_,
                        "type": "subscribe",
                        "payload": {"query": query, "variables": variables},
                    }
                )
            )
        else:
            start_message: ws_types.StartMessage = {
                "type": "start",
                "id": id_,
                "payload": {
                    "query": query,
                },
            }

            if variables is not None:
                start_message["payload"]["variables"] = variables

            await self.send_json_to(start_message)

        while True:
            message: transport_ws_types.Message = await self.receive_json_from(
                timeout=5
            )
            if message["type"] == "next":
                payload = message["payload"]
                ret = ExecutionResult(payload.get("data"), None)
                if "errors" in payload:
                    ret.errors = self.process_errors(payload.get("errors") or [])
                ret.extensions = payload.get("extensions", None)
                yield ret
            elif message["type"] == "error":
                error_payload = message["payload"]
                yield ExecutionResult(
                    data=None, errors=self.process_errors(error_payload)
                )
                return  # an error message is the last message for a subscription
            else:
                return

    def process_errors(self, errors: list[GraphQLFormattedError]) -> list[GraphQLError]:
        """Reconstructs a GraphQLError from a FormattedGraphQLError."""
        result = []
        for f_error in errors:
            error = GraphQLError(
                message=f_error["message"],
                extensions=f_error.get("extensions", None),
            )
            error.path = f_error.get("path", None)
            result.append(error)
        return result


__all__ = ["GraphQLWebsocketCommunicator"]



from typing import Any, Union

from starlette.background import BackgroundTasks
from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket

CustomContext = Union["BaseContext", dict[str, Any]]
MergedContext = Union[
    "BaseContext", dict[str, Any | BackgroundTasks | Request | Response | WebSocket]
]


class BaseContext:
    connection_params: Any | None = None

    def __init__(self) -> None:
        self.request: Request | WebSocket | None = None
        self.background_tasks: BackgroundTasks | None = None
        self.response: Response | None = None


__all__ = ["BaseContext"]



class NonTextMessageReceived(Exception):
    pass


class NonJsonMessageReceived(Exception):
    pass


class WebSocketDisconnected(Exception):
    pass


__all__ = [
    "NonJsonMessageReceived",
    "NonTextMessageReceived",
    "WebSocketDisconnected",
]



import datetime
import decimal
import uuid
from collections.abc import Callable
from operator import methodcaller

import dateutil.parser
from graphql import GraphQLError

from strawberry.types.scalar import ScalarDefinition


def wrap_parser(
    parser: Callable[[str], object],
    type_: str,
    exceptions: tuple[type[Exception], ...] = (ValueError,),
    include_error: bool = True,
    accept_non_string: bool = False,
) -> Callable[[object], object]:
    """Wrap a string parser so any invalid input becomes a clean coercion error.

    Non-string input is rejected up front so the parser can never crash with
    ``AttributeError``/``TypeError`` on unexpected value types and surface as
    a server-side error. ``accept_non_string`` stringifies non-string input
    instead of rejecting it, for scalars that accept it by design (``Decimal``
    accepts numeric input this way).
    """

    def inner(value: object) -> object:
        if not isinstance(value, str):
            if not accept_non_string:
                raise GraphQLError(
                    f'Value cannot represent a {type_}: "{value}". Expected a string.'
                )
            value = str(value)
        try:
            return parser(value)
        except exceptions as e:
            detail = f" {e}" if include_error else ""
            raise GraphQLError(
                f'Value cannot represent a {type_}: "{value}".{detail}'
            ) from None

    return inner


isoformat = methodcaller("isoformat")


DateDefinition: ScalarDefinition = ScalarDefinition(
    name="Date",
    description="Date (isoformat)",
    specified_by_url=None,
    serialize=isoformat,
    parse_value=wrap_parser(datetime.date.fromisoformat, "Date"),
    parse_literal=None,
    origin=datetime.date,
)

DateTimeDefinition: ScalarDefinition = ScalarDefinition(
    name="DateTime",
    description="Date with time (isoformat)",
    specified_by_url=None,
    serialize=isoformat,
    parse_value=wrap_parser(dateutil.parser.isoparse, "DateTime"),
    parse_literal=None,
    origin=datetime.datetime,
)

TimeDefinition: ScalarDefinition = ScalarDefinition(
    name="Time",
    description="Time (isoformat)",
    specified_by_url=None,
    serialize=isoformat,
    parse_value=wrap_parser(datetime.time.fromisoformat, "Time"),
    parse_literal=None,
    origin=datetime.time,
)

DecimalDefinition: ScalarDefinition = ScalarDefinition(
    name="Decimal",
    description="Decimal (fixed-point)",
    specified_by_url=None,
    serialize=str,
    parse_value=wrap_parser(
        decimal.Decimal,
        "Decimal",
        exceptions=(decimal.DecimalException,),
        include_error=False,
        accept_non_string=True,
    ),
    parse_literal=None,
    origin=decimal.Decimal,
)

UUIDDefinition: ScalarDefinition = ScalarDefinition(
    name="UUID",
    description=None,
    specified_by_url=None,
    serialize=str,
    parse_value=wrap_parser(uuid.UUID, "UUID"),
    parse_literal=None,
    origin=uuid.UUID,
)


def _verify_void(x: None) -> None:
    if x is not None:
        raise ValueError(f"Expected 'None', got '{x}'")


VoidDefinition: ScalarDefinition = ScalarDefinition(
    name="Void",
    description="Represents NULL values",
    specified_by_url=None,
    serialize=_verify_void,
    parse_value=_verify_void,
    parse_literal=None,
    origin=type(None),
)


__all__ = [
    "DateDefinition",
    "DateTimeDefinition",
    "DecimalDefinition",
    "TimeDefinition",
    "UUIDDefinition",
    "VoidDefinition",
]



from collections.abc import AsyncGenerator
from typing import Annotated, TypeVar


class StrawberryStreamable: ...


T = TypeVar("T")

Streamable = Annotated[AsyncGenerator[T, None], StrawberryStreamable()]
"""Represents a list that can be streamed using @stream.

Example:

```python
import strawberry


@strawberry.type
class Comment:
    id: strawberry.ID
    content: str


@strawberry.type
class Article:
    @strawberry.field
    @staticmethod
    async def comments() -> strawberry.Streamable[Comment]:
        for comment in fetch_comments():
            yield comment
```
"""

__all__ = ["Streamable"]



import os
from typing import Any

from starlette.applications import Starlette
from starlette.middleware.cors import CORSMiddleware
from starlette.routing import BaseRoute, Route, WebSocketRoute

from strawberry import Schema
from strawberry.asgi import GraphQL
from strawberry.cli.constants import DEV_SERVER_SCHEMA_ENV_VAR_KEY
from strawberry.utils.importer import import_module_symbol

schema_import_string = os.environ[DEV_SERVER_SCHEMA_ENV_VAR_KEY]
schema_symbol = import_module_symbol(schema_import_string, default_symbol_name="schema")

assert isinstance(schema_symbol, Schema)
graphql_app = GraphQL[Any, Any](schema_symbol)

paths = ["/", "/graphql"]
routes: list[BaseRoute] = []
for path in paths:
    routes.append(Route(path, graphql_app))
    routes.append(WebSocketRoute(path, graphql_app))

app = Starlette(debug=True, routes=routes)
app.add_middleware(
    CORSMiddleware, allow_headers=["*"], allow_origins=["*"], allow_methods=["*"]
)



from pathlib import Path

import typer

from strawberry.cli.app import app
from strawberry.schema_codegen import codegen


@app.command(help="Generate code from a query")
def schema_codegen(
    schema: Path = typer.Argument(exists=True),
    output: Path | None = typer.Option(
        None,
        "-o",
        "--output",
        file_okay=True,
        dir_okay=False,
        writable=True,
        resolve_path=True,
    ),
) -> None:
    generated_output = codegen(schema.read_text())

    if output is None:
        typer.echo(generated_output)
        return

    output.parent.mkdir(parents=True, exist_ok=True)
    output.write_text(generated_output)

    typer.echo(f"Code generated at `{output.name}`")



from __future__ import annotations

import dataclasses
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    overload,
)
from typing_extensions import Unpack

from strawberry.types.field import field as base_field

from .params import FederationFieldParams, process_federation_field_directives

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence
    from typing import Literal

    from strawberry.extensions.field_extension import FieldExtension
    from strawberry.permission import BasePermission
    from strawberry.types.field import (
        _RESOLVER_TYPE,
        _RESOLVER_TYPE_ASYNC,
        _RESOLVER_TYPE_SYNC,
    )

T = TypeVar("T")

# NOTE: we are separating the sync and async resolvers because using both
# in the same function will cause mypy to raise an error. Not sure if it is a bug


@overload
def field(
    *,
    resolver: _RESOLVER_TYPE_ASYNC[T],
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    init: Literal[False] = False,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
    **federation_kwargs: Unpack[FederationFieldParams],
) -> T: ...


@overload
def field(
    *,
    resolver: _RESOLVER_TYPE_SYNC[T],
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    init: Literal[False] = False,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
    **federation_kwargs: Unpack[FederationFieldParams],
) -> T: ...


@overload
def field(
    *,
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    init: Literal[True] = True,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
    **federation_kwargs: Unpack[FederationFieldParams],
) -> Any: ...


def field(
    resolver: _RESOLVER_TYPE[Any] | None = None,
    *,
    name: str | None = None,
    is_subscription: bool = False,
    description: str | None = None,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
    # This init parameter is used by PyRight to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False] | None = None,
    **federation_kwargs: Unpack[FederationFieldParams],
) -> Any:
    directives = process_federation_field_directives(directives, **federation_kwargs)

    return base_field(  # type: ignore
        resolver=resolver,  # type: ignore
        name=name,
        is_subscription=is_subscription,
        description=description,
        permission_classes=permission_classes,
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        init=init,  # type: ignore
        directives=directives,
        metadata=metadata,
        extensions=extensions,
        graphql_type=graphql_type,
    )


__all__ = ["field"]



from graphql.version import VersionInfo, version_info

IS_GQL_33 = version_info >= VersionInfo.from_str("3.3.0a0")
IS_GQL_32 = not IS_GQL_33

if IS_GQL_33 and version_info < VersionInfo.from_str("3.3.0a12"):
    raise ImportError(
        f"graphql-core {version_info} is not supported. "
        "Please upgrade to graphql-core >= 3.3.0a12."
    )



from __future__ import annotations

from typing import Any

import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.extensions.field_extension import FieldExtension
from strawberry.types.arguments import StrawberryArgument
from strawberry.types.field import StrawberryField
from strawberry.utils.str_converters import capitalize_first, to_camel_case


class InputMutationExtension(FieldExtension):
    def apply(self, field: StrawberryField) -> None:
        resolver = field.base_resolver
        assert resolver

        name = field.graphql_name or to_camel_case(resolver.name)
        type_dict: dict[str, Any] = {
            "__doc__": f"Input data for `{name}` mutation",
            "__annotations__": {},
        }
        annotations = resolver.wrapped_func.__annotations__

        for arg in resolver.arguments:
            arg_field = StrawberryField(
                python_name=arg.python_name,
                graphql_name=arg.graphql_name,
                description=arg.description,
                deprecation_reason=arg.deprecation_reason,
                default=arg.default,
                type_annotation=arg.type_annotation,
                directives=tuple(arg.directives),
            )
            type_dict[arg_field.python_name] = arg_field
            type_dict["__annotations__"][arg_field.python_name] = annotations[
                arg.python_name
            ]

        caps_name = capitalize_first(name)
        new_type = strawberry.input(type(f"{caps_name}Input", (), type_dict))
        field.arguments = [
            StrawberryArgument(
                python_name="input",
                graphql_name=None,
                type_annotation=StrawberryAnnotation(
                    new_type,
                    namespace=resolver._namespace,
                ),
                description=type_dict["__doc__"],
            )
        ]

    def map_arguments(self, kwargs: dict[str, Any]) -> dict[str, Any]:
        # Unpack the generated ``input`` object into the resolver's individual
        # keyword arguments, so the resolver keeps its original signature.
        input_args = kwargs.pop("input")
        return {**kwargs, **vars(input_args)}


__all__ = ["InputMutationExtension"]



from .base import get_object_definition, has_object_definition
from .execution import (
    ExecutionContext,
    ExecutionResult,
    StreamExecutionResult,
    SubscriptionExecutionResult,
)
from .info import Info

__all__ = [
    "ExecutionContext",
    "ExecutionResult",
    "Info",
    "Info",
    "StreamExecutionResult",
    "SubscriptionExecutionResult",
    "get_object_definition",
    "has_object_definition",
]



from typing import Any
from typing_extensions import Self


class UnsetType:
    __instance: Self | None = None

    def __new__(cls: type[Self]) -> Self:
        if cls.__instance is None:
            ret = super().__new__(cls)
            cls.__instance = ret
            return ret
        return cls.__instance

    def __str__(self) -> str:
        return ""

    def __repr__(self) -> str:
        return "UNSET"

    def __bool__(self) -> bool:
        return False


UNSET: Any = UnsetType()
"""A special value that can be used to represent an unset value in a field or argument.
Similar to `undefined` in JavaScript, this value can be used to differentiate between
a field that was not set and a field that was set to `None` or `null`.

Example:

```python
import strawberry


@strawberry.input
class UserInput:
    name: str | None = strawberry.UNSET
    age: int | None = strawberry.UNSET
```

In the example above, if `name` or `age` are not provided when creating a `UserInput`
object, they will be set to `UNSET` instead of `None`. Use `is UNSET` to check
whether a value is unset.
"""

__all__ = [
    "UNSET",
]
